import asyncio
from threading import Thread, Event
from time import sleep

from dbus_next.aio import MessageBus as DbusMessageBus
from dbus_next.constants import BusType
from dbus_next.message import Message as DbusMessage, \
    MessageType as DbusMessageType
from dbus_next.service import ServiceInterface, method, dbus_property, PropertyAccess
from ovos_bus_client.message import Message
from ovos_utils.log import LOG
from ovos_utils.ocp import TrackState, PlaybackType, PlayerState, LoopState


class MprisPlayerCtl(Thread):
    """ detects other media players in the system and integrates with them
    - stop internal playback when an external player starts
    - display gui with data from external media player
    - provide voice control over the external player
    - provide gui controls over the external player
    - advertises OCP over mpris so external applications can control it
        eg, KDE connect will allow controlling OCP via the phone
    """

    def __init__(self, daemonic=True, manage_players=False):
        super(MprisPlayerCtl, self).__init__()
        self.dbus = None
        self.loop = asyncio.get_event_loop()

        self.setDaemon(daemonic)
        self.shutdown_event = Event()
        self.stop_event = Event()
        self.pause_event = Event()
        self.resume_event = Event()
        self.next_event = Event()
        self.prev_event = Event()

        self.mediaPlayer2Interface = _MediaPlayer2Interface('org.mpris.MediaPlayer2')
        self.mediaPlayer2PlayerInterface = _MediaPlayer2PlayerInterface('org.mpris.MediaPlayer2.Player')
        self._ocp_player = None

        self.main_player = None
        self.players = {}
        self.player_meta = {}
        self._player_fails = {}
        self.manage_players = manage_players
        # TODO from .conf
        self.ignored_players = [
            "org.mpris.MediaPlayer2.OCP",
            "org.mpris.MediaPlayer2.plasma-browser-integration"  # browsers already show up as individual players
        ]

    def bind(self, ocp_player):
        self._ocp_player = ocp_player
        self.mediaPlayer2PlayerInterface.bind(self._ocp_player)
        self.mediaPlayer2Interface.bind(self._ocp_player)
        self.start()

    @property
    def dbus_type(self):
        if self._ocp_player:
            config = self._ocp_player.settings.get("dbus_type") or "session"
            return BusType.SYSTEM if config.lower().strip() == "system" else \
                BusType.SESSION
        return BusType.SESSION

    async def export_ocp(self):
        self.dbus.export('/org/mpris/MediaPlayer2', self.mediaPlayer2Interface)
        self.dbus.export('/org/mpris/MediaPlayer2', self.mediaPlayer2PlayerInterface)
        await self.dbus.request_name('org.mpris.MediaPlayer2.OCP')

    def update_props(self, props):
        self.mediaPlayer2PlayerInterface.emit_properties_changed(props)

    def _update_ocp(self):
        if self.stop_event.is_set() or not self.manage_players:
            return
        if self._ocp_player and self.player_meta.get(self.main_player):
            data = self.player_meta[self.main_player]

            # reset ocp, it will display metadata of current track
            if self._ocp_player.active_skill != self.main_player:
                self._ocp_player.reset()
                self._ocp_player.gui.show_player()

            # player state
            state = data.get("state") or "Playing"
            if state == "Paused":
                self._ocp_player.set_player_state(PlayerState.PAUSED)
            elif state == "Playing":
                self._ocp_player.set_player_state(PlayerState.PLAYING)
            else:
                self._ocp_player.set_player_state(PlayerState.STOPPED)
            self._ocp_player.loop_state = data.get("loop_state") or \
                                          self._ocp_player.loop_state
            self._ocp_player.shuffle = data.get("shuffle") or \
                                       self._ocp_player.shuffle

            # update ocp metadata
            data["skill_id"] = data["external_player"]
            data["bg_image"] = data.get("image")
            data["playback"] = PlaybackType.MPRIS
            data["status"] = TrackState.PLAYING_MPRIS
            self._ocp_player.set_now_playing(data)

    async def handle_new_player(self, data):
        if data['name'] not in self._player_fails:
            LOG.info(f"Found MPRIS Player: {data['name']}")

    async def handle_player_shuffle(self, shuffle):
        LOG.info(f"MPRIS Player Shuffle: {shuffle}")

    async def handle_player_loop_state(self, state):
        LOG.info(f"MPRIS Player Repeat: {state}")

    async def handle_player_state(self, state):
        LOG.info(f"MPRIS Player State: {state}")

    async def handle_lost_player(self, name):
        LOG.info(f"Lost MPRIS Player: {name}")
        if name in self.player_meta:
            self.player_meta.pop(name)
        if name in self.players:
            self.players.pop(name)

    async def handle_sync_player(self, data):
        if data.get("state") == 'Playing':
            await self._set_main_player(data["external_player"])
        elif data["external_player"] == self.main_player:
            self._update_ocp()

    async def _set_main_player(self, name):
        self.main_player = name
        # if there are multiple external players playing, stop the
        # previous ones!
        if self.manage_players:
            self._update_ocp()
            for p in self.players:
                if p != name:
                    await self._stop_player(p)

    async def _play_prev(self, name, max_tries=1):
        if name not in self.players:
            LOG.error(f"Invalid player: {name}")
            return
        try:
            if self.player_meta[name]["state"] == "Playing":
                LOG.debug(f"player previous {name}")
                player = self.players[name].get_interface(
                    'org.mpris.MediaPlayer2.Player')
                await player.call_previous()
        except:
            max_tries -= 1
            if max_tries > 0:
                await self._play_prev(name, max_tries)
            else:
                LOG.warning(f"player {name} does not support Previous")

    async def _play_next(self, name, max_tries=1):
        if name not in self.players:
            LOG.error(f"Invalid player: {name}")
            return
        try:
            if self.player_meta[name]["state"] == "Playing":
                LOG.debug(f"player next {name}")
                player = self.players[name].get_interface(
                    'org.mpris.MediaPlayer2.Player')
                await player.call_next()
        except:
            max_tries -= 1
            if max_tries > 0:
                await self._play_next(name, max_tries)
            else:
                LOG.warning(f"player {name} does not support Next")

    async def _pause_player(self, name, max_tries=1):
        if name not in self.players:
            LOG.error(f"Invalid player: {name}")
            return
        try:
            if self.player_meta[name]["state"] == "Playing":
                LOG.debug(f"pausing player {name}")
                player = self.players[name].get_interface(
                    'org.mpris.MediaPlayer2.Player')
                await player.call_pause()
        except:
            max_tries -= 1
            if max_tries > 0:
                await self._pause_player(name, max_tries)
            else:
                LOG.warning(f"player {name} can not be paused")

    async def _resume_player(self, name, max_tries=1):
        if name not in self.players:
            LOG.error(f"Invalid player: {name}")
            return
        try:
            if self.player_meta[name]["state"] != "Playing":
                LOG.debug(f"resuming player {name}")
                player = self.players[name].get_interface(
                    'org.mpris.MediaPlayer2.Player')
                await player.call_play()
        except:
            max_tries -= 1
            if max_tries > 0:
                await self._resume_player(name, max_tries)
            else:
                LOG.warning(f"player {name} can not be resumed")

    async def _stop_player(self, name, max_tries=1):
        if name not in self.players:
            LOG.error(f"Invalid player: {name}")
            return
        try:
            if self.player_meta[name]["state"] == "Playing":
                LOG.debug(f"stopping player {name}")
                player = self.players[name].get_interface(
                    'org.mpris.MediaPlayer2.Player')
                await player.call_stop()
        except:
            max_tries -= 1
            if max_tries > 0:
                await self._stop_player(name, max_tries)
            else:
                LOG.warning(f"player {name} can not be stopped")
        if name == self.main_player:
            self.main_player = None

    async def _stop_all(self):
        for p in self.players:
            await self._stop_player(p)

    async def _pause_all(self):
        for p in self.players:
            await self._pause_player(p)

    async def scan_players(self):
        reply = await self.dbus.call(
            DbusMessage(destination='org.freedesktop.DBus',
                        path='/org/freedesktop/DBus',
                        interface='org.freedesktop.DBus',
                        member='ListNames'))

        if reply.message_type == DbusMessageType.ERROR:
            raise Exception(reply.body[0])

        players = []
        for name in reply.body[0]:
            if "org.mpris.MediaPlayer2" in name:
                if name in self.players or name in self.ignored_players:
                    continue
                await self.handle_new_player({"name": name})
                introspection = await self.dbus.introspect(
                    name, '/org/mpris/MediaPlayer2')
                self.players[name] = self.dbus.get_proxy_object(
                    name, '/org/mpris/MediaPlayer2', introspection)
                self._create_player_handler(name)
                await self.query_player(name)
        return players

    def _create_player_handler(self, name):
        player = self.players[name]
        try:
            properties = player.get_interface(
                'org.freedesktop.DBus.Properties')
        except:
            # chromium
            LOG.warning(f"Player {name} does not allow reading properties")
            return

        # listen to signals
        async def on_properties_changed(interface_name,
                                        changed_properties,
                                        invalidated_properties):
            for changed, variant in changed_properties.items():
                player_name = properties.bus_name
                if player_name in self.ignored_players:
                    continue
                if changed == "PlaybackStatus":
                    await self.handle_player_state(variant.value)
                    state = self.player_meta[player_name].get("state")
                    if state != variant.value or not state:
                        self.player_meta[player_name]["state"] = variant.value
                        await self.handle_sync_player(
                            {"state": variant.value,
                             "external_player": player_name})
                elif changed == "Metadata":
                    await self.update_player_meta(player_name, variant.value)
                elif changed == "Shuffle":
                    self.player_meta[player_name]["shuffle"] = variant.value
                    await self.handle_player_shuffle(variant.value)
                elif changed == "LoopStatus":
                    if variant.value == "Track":
                        state = LoopState.REPEAT_TRACK
                    elif variant.value == "Playlist":
                        state = LoopState.REPEAT
                    else:
                        state = LoopState.NONE
                    self.player_meta[player_name]["loop_state"] = state
                    await self.handle_player_loop_state(state)
                # else:
                #    LOG.debug(f'{changed} - {variant.value}')

        properties.on_properties_changed(on_properties_changed)

    async def update_player_meta(self, name, meta):
        ocp_data = {"external_player": name}

        # these are injected when player is queried
        ocp_data["state"] = meta.get("state")
        ocp_data["loop_state"] = meta.get("loop_state")

        for k, v in meta.items():
            if k == "xesam:title":
                ocp_data["title"] = v.value
            elif k == "xesam:artist":
                ocp_data["artist"] = v.value[0]
            elif k == "xesam:album":
                ocp_data["album"] = v.value
            elif k == "mpris:artUrl":
                ocp_data["image"] = v.value
            elif k == "mpris:length":
                ocp_data["length"] = v.value

        self.player_meta[name] = ocp_data
        await self.handle_sync_player(ocp_data)

    async def query_player(self, name):
        if self._player_fails.get(name, 0) >= 3:
            # do not keep querying players that dont expose full mpris functionality
            return
        if name not in self.players:
            LOG.error(f"Invalid player: {name}")
            return
        try:
            player = self.players[name].get_interface(
                'org.mpris.MediaPlayer2.Player')
            meta = await player.get_metadata()
            meta["external_player"] = name
            try:
                meta["state"] = await player.get_playback_status()
            except:  # dbus_next.errors.DBusError
                pass
            try:
                loop_status = await player.get_loop_status()
                if loop_status == "None":
                    # The playback will stop when there are no more tracks to play
                    meta["loop_state"] = LoopState.NONE
                elif loop_status == "Track":
                    # The current track will start again from the begining once it has finished playing
                    meta["loop_state"] = LoopState.REPEAT_TRACK
                elif loop_status == "Playlist":
                    # The playback loops through a list of tracks
                    meta["loop_state"] = LoopState.REPEAT
            except AttributeError:
                pass  # not all players expose this
            await self.update_player_meta(name, meta)
            self._player_fails[name] = 0
        except Exception as e:  # chromium / player closed
            if name not in self._player_fails:
                self._player_fails[name] = 0
            self._player_fails[name] += 1
            if self._player_fails[name] > 3:
                LOG.debug(f"failed to query player {name}")
                await self.handle_lost_player(name)

    async def event_loop(self):
        self.shutdown_event.clear()
        self.stop_event.clear()
        self.pause_event.clear()

        while not self.shutdown_event.is_set():
            if not self._ocp_player:
                # wait for self.bind to be called
                sleep(1)
                continue

            if not self.dbus:
                self.dbus = await DbusMessageBus(
                    bus_type=self.dbus_type).connect()
                await self.export_ocp()

            # ocp requests to manipulate external players
            if self.stop_event.is_set():
                await self._stop_all()
                self.stop_event.clear()

            if self.pause_event.is_set():
                await self._pause_all()
                self.pause_event.clear()

            if self.prev_event.is_set():
                await self._play_prev(self.main_player)
                self.prev_event.clear()

            if self.next_event.is_set():
                await self._play_next(self.main_player)
                self.next_event.clear()

            if self.resume_event.is_set():
                await self._resume_player(self.main_player)
                self.resume_event.clear()

            # scan for new external players
            await self.scan_players()
            sleep(1)  # TODO configurable time between checks

            # sync player meta, not all players send all events properly...
            # eg, firefox videos do not send events if they autoplay, only if
            # you click the play button
            for player in list(self.players.keys()):
                await self.query_player(player)
            sleep(1)  # TODO configurable time between checks

    def run(self):
        try:
            self.loop.run_until_complete(self.event_loop())
        except Exception as e:
            LOG.error(e)

    def play_prev(self):
        self.prev_event.set()

    def play_next(self):
        self.next_event.set()

    def resume(self):
        self.resume_event.set()

    def pause(self):
        self.pause_event.set()

    def stop(self):
        self.stop_event.set()

    def shutdown(self):
        self.stop()
        self.shutdown_event.set()
        self.loop.stop()
        while self.loop.is_running():
            sleep(0.2)
        self.loop.close()


class _MediaPlayer2Interface(ServiceInterface):
    def __init__(self, name='org.mpris.MediaPlayer2'):
        super().__init__(name)
        self._ocp_player = None
        self._identity = "OCP"
        self._desktopEntry = "OCP"
        self._supportedMimeTypes = ["audio/mpeg", "audio/x-mpeg", "video/mpeg", "video/x-mpeg", "video/mpeg-system",
                                    "video/x-mpeg-system", "video/mp4", "audio/mp4", "video/x-msvideo",
                                    "video/quicktime", "application/ogg", "application/x-ogg", "video/x-ms-asf",
                                    "video/x-ms-asf-plugin", "application/x-mplayer2", "video/x-ms-wmv",
                                    "video/x-google-vlc-plugin", "audio/wav", "audio/x-wav", "audio/3gpp", "video/3gpp",
                                    "audio/3gpp2", "video/3gpp2", "video/divx", "video/flv", "video/x-flv",
                                    "video/x-matroska", "audio/x-matroska", "application/xspf+xml"]
        self._supportedUriSchemes = ["file", "http", "https", "rtsp", "realrtsp", "pnm", "ftp", "mtp", "smb", "mms",
                                     "mmsu", "mmst", "mmsh", "unsv", "itpc", "icyx", "rtmp", "rtp", "dccp", "dvd",
                                     "vcd"]
        self._canQuit = False
        self._hasTrackList = False

    def bind(self, ocp_player):
        self._ocp_player = ocp_player
        self._hasTrackList = len(self._ocp_player.playlist) > 0

    def update_props(self, props):
        self.emit_properties_changed(props)

    @dbus_property(access=PropertyAccess.READ)
    def Identity(self) -> 's':
        return self._identity

    @dbus_property(access=PropertyAccess.READ)
    def DesktopEntry(self) -> 's':
        return self._desktopEntry

    @dbus_property(access=PropertyAccess.READ)
    def SupportedMimeTypes(self) -> 'as':
        return self._supportedMimeTypes

    @dbus_property(access=PropertyAccess.READ)
    def SupportedUriSchemes(self) -> 'as':
        return self._supportedUriSchemes

    @dbus_property(access=PropertyAccess.READ)
    def HasTrackList(self) -> 'b':
        return True

    @dbus_property(access=PropertyAccess.READ)
    def CanQuit(self) -> 'b':
        return self._canQuit

    @dbus_property(access=PropertyAccess.READ)
    def CanSetFullscreen(self) -> 'b':
        return False

    @dbus_property(access=PropertyAccess.READ)
    def Fullscreen(self) -> 'b':
        return False

    @dbus_property(access=PropertyAccess.READ)
    def CanRaise(self) -> 'b':
        return False

    @method()
    def Quit(self):
        if self._canQuit:
            self._ocp_player.shutdown()


class _MediaPlayer2PlayerInterface(ServiceInterface):
    def __init__(self, name):
        super().__init__(name)
        self._ocp_player = None

    def bind(self, ocp_player):
        self._ocp_player = ocp_player

    @dbus_property(access=PropertyAccess.READ)
    def Metadata(self) -> 'a{sv}':
        return self._ocp_player.now_playing.mpris_metadata

    @dbus_property(access=PropertyAccess.READ)
    def PlaybackStatus(self) -> 's':
        # TODO validate strings
        if self._ocp_player.state == PlayerState.PLAYING:
            return "Playing"
        if self._ocp_player.state == PlayerState.PAUSED:
            return "Paused"
        return "Stopped"

    @dbus_property()
    def LoopStatus(self) -> 's':
        # TODO validate strings
        if self._ocp_player.loop_state == LoopState.REPEAT_TRACK:
            return "RepeatTrack"
        if self._ocp_player.loop_state == LoopState.REPEAT:
            return "Repeat"
        return "None"

    @LoopStatus.setter
    def LoopStatus_setter(self, val: 's'):
        # TODO translate state
        self._ocp_player.loop_state = val

    @dbus_property()
    def Shuffle(self) -> 'b':
        return self._ocp_player.shuffle

    @Shuffle.setter
    def Shuffle_setter(self, val: 'b'):
        self._ocp_player.shuffle = val

    @dbus_property()
    def Volume(self) -> 'd':
        msg = self._ocp_player.bus.wait_for_response(Message("mycroft.volume.get"), timeout=0.5)
        if msg:
            return float(msg.data["percent"])
        return 1.0

    @Volume.setter
    def Volume_setter(self, val: 'd'):
        self._ocp_player.bus.emit(Message("mycroft.volume.set", {"percent": val}))

    @dbus_property(access=PropertyAccess.READ)
    def Rate(self) -> 'd':
        return 1

    @dbus_property(access=PropertyAccess.READ)
    def Position(self) -> 'd':
        return 1  # TODO from ocp_player

    @dbus_property(access=PropertyAccess.READ)
    def CanPlay(self) -> 'b':
        return self._ocp_player.state == PlayerState.PAUSED

    @dbus_property(access=PropertyAccess.READ)
    def CanPause(self) -> 'b':
        return self._ocp_player.state == PlayerState.PLAYING

    @dbus_property(access=PropertyAccess.READ)
    def CanSeek(self) -> 'b':
        return False

    @dbus_property(access=PropertyAccess.READ)
    def CanGoNext(self) -> 'b':
        return self._ocp_player.can_next

    @dbus_property(access=PropertyAccess.READ)
    def CanGoPrevious(self) -> 'b':
        return self._ocp_player.can_prev

    @dbus_property(access=PropertyAccess.READ)
    def CanControl(self) -> 'b':
        return True

    @method()
    def Previous(self):
        self._ocp_player.play_prev()

    @method()
    def Next(self):
        self._ocp_player.play_next()

    @method()
    def Stop(self):
        self._ocp_player.pause()

    @method()
    def Play(self):
        self._ocp_player.resume()

    @method()
    def Pause(self):
        self._ocp_player.pause()

    @method()
    def PlayPause(self):
        if self._ocp_player.state == PlayerState.PAUSED:
            self._ocp_player.resume()
        else:
            self._ocp_player.pause()



import shutil
from functools import wraps
from os import makedirs
from os.path import expanduser, isfile, join, dirname, exists

from ovos_bus_client.session import SessionManager
from ovos_plugin_manager.ocp import load_stream_extractors
from ovos_utils.log import LOG


def require_default_session():
    def _decorator(func):
        @wraps(func)
        def func_wrapper(self, message=None):
            sess = SessionManager.get(message)
            validated = message is None or \
                        not self.validate_source or \
                        sess.session_id == "default"
            if validated:
                return func(self, message)
            LOG.debug(f"ignoring OCP bus message, not from 'default' session': {message.context if message else sess.session_id}")
            return None

        return func_wrapper

    return _decorator


def ocp_plugins():
    return load_stream_extractors()


def is_qtav_available():
    return exists("/usr/include/qt/QtAV") or \
        exists("/usr/lib/qt/qml/QtAV") or \
        exists("/usr/lib/libQtAV.so")


def extract_metadata(uri):
    # backwards compat
    from ovos_ocp_files_plugin.plugin import OCPFilesMetadataExtractor
    return OCPFilesMetadataExtractor.extract_metadata(uri)


def create_desktop_file():
    res = join(dirname(__file__), "res", "desktop")
    desktop_path = expanduser("~/.local/share/applications")
    icon_path = expanduser("~/.local/share/icons")
    makedirs(desktop_path, exist_ok=True)
    makedirs(icon_path, exist_ok=True)

    src_desktop = join(res, "OCP.desktop")
    dst_desktop = join(desktop_path, "OCP.desktop")
    if not isfile(dst_desktop):
        shutil.copy(src_desktop, dst_desktop)

    src_icon = join(res, "OCP.png")
    dst_icon = join(icon_path, "OCP.png")
    if not isfile(dst_icon):
        shutil.copy(src_icon, dst_icon)



import random
from os.path import join, dirname
from typing import List, Union, Optional

from ovos_bus_client.apis.ocp import ClassicAudioServiceInterface
from ovos_bus_client.message import Message, dig_for_message
from ovos_utils.log import LOG
from ovos_utils.ocp import (PluginStream, LoopState, MediaState, PlayerState, TrackState, dict2entry,
                            PlaybackType, MediaEntry, PlaybackMode, Playlist)
from ovos_workshop.app import OVOSAbstractApplication

from ovos_config import Configuration
from ovos_plugin_common_play.ocp.constants import OCP_ID
from ovos_plugin_common_play.ocp.gui import OCPMediaPlayerGUI
from ovos_plugin_common_play.ocp.media import NowPlaying
from ovos_plugin_common_play.ocp.mpris import MprisPlayerCtl
from ovos_plugin_common_play.ocp.search import OCPSearch
from ovos_plugin_common_play.ocp.utils import require_default_session


class OCPMediaPlayer(OVOSAbstractApplication):
    def __init__(self, bus=None, settings=None, lang=None, gui=None,
                 resources_dir=None, skill_id=OCP_ID, validate_source=True,
                 native_sources: Optional[List[str]] = None,
                 **kwargs):
        resources_dir = resources_dir or join(dirname(__file__), "res")
        gui = gui or OCPMediaPlayerGUI(bus=bus)

        self.validate_source = validate_source
        self.native_sources = native_sources or Configuration()["Audio"].\
            get("native_sources", ["debug_cli", "audio"])
        # Define things referenced in `bind`
        self.now_playing: NowPlaying = NowPlaying()
        self.media: OCPSearch = OCPSearch()
        self.state: PlayerState = PlayerState.STOPPED
        self.loop_state: LoopState = LoopState.NONE
        self.media_state: MediaState = MediaState.NO_MEDIA
        self.playlist: Playlist = Playlist()
        self.shuffle: bool = False
        self.audio_service = None
        self._audio_backend = None
        self.track_history = {}  # Dict of track URI to play count

        super().__init__(skill_id=skill_id, bus=bus, gui=gui,
                         resources_dir=resources_dir, **kwargs)
        if settings:
            self.settings.merge(settings)

        self._paused_on_duck = False

        # mpris settings
        manage_players = self.settings.get("manage_external_players", False)
        if self.settings.get('disable_mpris'):
            LOG.info("MPRIS integration is disabled")
            self.mpris = None
        else:
            self.mpris = MprisPlayerCtl(manage_players=manage_players)
            self.mpris.bind(self)

    def bind(self, bus=None):
        """
        Initialize components that need a MessageBusClient or instance of this
        object.
        @param bus: MessageBusClient object to register events on
        """
        super(OCPMediaPlayer, self).bind(bus)
        self.now_playing.bind(self)
        self.media.bind(self)
        self.gui.bind(self)
        self.audio_service = ClassicAudioServiceInterface(self.bus)
        self.register_bus_handlers()

    def register_bus_handlers(self):
        # audio ducking TODO improve to wait for end of speech ?
        self.add_event('recognizer_loop:record_begin',
                       self.handle_duck_request)
        self.add_event('recognizer_loop:record_end',
                       self.handle_unduck_request)

        # mycroft-gui media service
        self.add_event('gui.player.media.service.sync.status',
                       self.handle_player_state_update)
        self.add_event("gui.player.media.service.get.next",
                       self.handle_next_request)
        self.add_event("gui.player.media.service.get.previous",
                       self.handle_prev_request)
        self.add_event("gui.player.media.service.get.repeat",
                       self.handle_repeat_toggle_request)
        self.add_event("gui.player.media.service.get.shuffle",
                       self.handle_shuffle_toggle_request)

        # ovos common play bus api
        self.add_event('ovos.common_play.player.status',
                       self.handle_state_request)
        self.add_event('ovos.common_play.player.state',
                       self.handle_player_state_update)
        self.add_event('ovos.common_play.media.state',
                       self.handle_player_media_update)
        self.add_event('ovos.common_play.play',
                       self.handle_play_request)
        self.add_event('ovos.common_play.pause',
                       self.handle_pause_request)
        self.add_event('ovos.common_play.resume',
                       self.handle_resume_request)
        self.add_event('ovos.common_play.stop',
                       self.handle_stop_request)
        self.add_event('ovos.common_play.next',
                       self.handle_next_request)
        self.add_event('ovos.common_play.previous',
                       self.handle_prev_request)
        self.add_event('ovos.common_play.seek',
                       self.handle_seek_request)
        self.add_event('ovos.common_play.get_track_length',
                       self.handle_track_length_request)
        self.add_event('ovos.common_play.set_track_position',
                       self.handle_set_track_position_request)
        self.add_event('ovos.common_play.get_track_position',
                       self.handle_track_position_request)
        self.add_event('ovos.common_play.track_info',
                       self.handle_track_info_request)
        self.add_event('ovos.common_play.list_backends',
                       self.handle_list_backends_request)
        self.add_event('ovos.common_play.playlist.set',
                       self.handle_playlist_set_request)
        self.add_event('ovos.common_play.playlist.clear',
                       self.handle_playlist_clear_request)
        self.add_event('ovos.common_play.playlist.queue',
                       self.handle_playlist_queue_request)
        self.add_event('ovos.common_play.duck',
                       self.handle_duck_request)
        self.add_event('ovos.common_play.unduck',
                       self.handle_unduck_request)
        self.add_event('ovos.common_play.shuffle.set',
                       self.handle_set_shuffle)
        self.add_event('ovos.common_play.shuffle.unset',
                       self.handle_unset_shuffle)
        self.add_event('ovos.common_play.repeat.set',
                       self.handle_set_repeat)
        self.add_event('ovos.common_play.repeat.unset',
                       self.handle_unset_repeat)
        self.add_event('ovos.common_play.search.populate',
                       self.handle_search_replace)

        # GUI Configuration Events
        self.add_event('ovos.common_play.gui.enable_app_timeout',
                       self.handle_enable_app_timeout)
        self.add_event('ovos.common_play.gui.set_app_timeout',
                       self.handle_set_app_timeout)
        self.add_event('ovos.common_play.gui.timeout.mode',
                       self.handle_set_app_timeout_mode)

    @require_default_session()
    def handle_state_request(self, message: Message):
        msg = message.response({
            "state": self.state,
            "media_state": self.media_state,
            "shuffle": self.shuffle,
            "repeat": self.loop_state.value
        })
        self.bus.emit(msg)

    @property
    def active_skill(self) -> str:
        """
        Return the skill_id of the skill providing the current media
        """
        return self.now_playing.skill_id

    @property
    def active_backend(self) -> PlaybackType:
        """
        Return the PlaybackType for the current media
        """
        return self.now_playing.playback

    @property
    def tracks(self) -> List[MediaEntry]:
        """
        Return the current queue as a list of MediaEntry objects
        """
        return self.playlist.entries

    @property
    def disambiguation(self) -> List[MediaEntry]:
        """
        Return a list of the previous search results as MediaEntry objects
        """
        return self.media.search_playlist.entries

    @property
    def can_prev(self) -> bool:
        """
        Return true if there is a previous track in the queue to skip to
        """
        if self.active_backend != PlaybackType.MPRIS and \
                self.playlist.is_first_track:
            return False
        return True

    @property
    def can_next(self) -> bool:
        """
        Return true if there is a next track in the queue to skip to
        """
        if self.loop_state != LoopState.NONE or \
                self.shuffle or \
                self.active_backend == PlaybackType.MPRIS:
            return True
        elif self.settings.get("merge_search", True) and \
                not self.media.search_playlist.is_last_track:
            return True
        elif not self.playlist.is_last_track:
            return True
        return False

    # state
    def set_media_state(self, state: MediaState):
        """
        Set self.media_state and emit an event announcing this state change.
        @param state: New MediaState
        """
        if not isinstance(state, MediaState):
            raise TypeError(f"Expected MediaState and got: {state}")
        if state == self.media_state:
            return
        self.media_state = state
        self.bus.emit(Message("ovos.common_play.media.state",
                              {"state": self.media_state}))

    def set_player_state(self, state: PlayerState):
        """
        Set self.state, update the GUI and MPRIS (if available), and emit an
        event announcing this state change.
        @param state: New PlayerState
        """
        if not isinstance(state, PlayerState):
            raise TypeError(f"Expected PlayerState and got: {state}")
        if state == self.state:
            return
        self.state = state
        state2str = {PlayerState.PLAYING: "Playing",
                     PlayerState.PAUSED: "Paused",
                     PlayerState.STOPPED: "Stopped"}
        self.gui["status"] = state2str[self.state]
        if self.mpris:
            self.mpris.update_props({"CanPause": self.state == PlayerState.PLAYING,
                                     "CanPlay": self.state == PlayerState.PAUSED,
                                     "PlaybackStatus": state2str[state]})
        self.bus.emit(Message("ovos.common_play.player.state",
                              {"state": self.state}))

    def set_now_playing(self, track: Union[dict, MediaEntry, Playlist, PluginStream]):
        """
        Set `track` as the currently playing media, update the playlist, and
        notify any GUI or MPRIS clients. Adds `track` to `playlist`
        @param track: MediaEntry or dict representation of a MediaEntry to play
        """
        LOG.debug(f"Playing: {track}")
        if isinstance(track, dict):
            LOG.debug(f"Handling dict track: {track}")
            if "uri" not in track:  # TODO handle this better
                track["uri"] = "external:"  # when syncing from MPRIS uri is missing
            track = dict2entry(track)
        if not isinstance(track, (MediaEntry, Playlist, PluginStream)):
            raise ValueError(f"Expected MediaEntry/Playlist, but got: {track}")

        try:
            idx = self.playlist.index(track)  # find the entry in "now playing"
        except ValueError:
            idx = -1
        if isinstance(track, PluginStream):
            track = track.extract_media_entry(video=track.playback == PlaybackType.VIDEO)
            LOG.info(f"PluginStream extracted: {track}")
            if idx >= 0:
                self.playlist[idx] = track  # update extracted plugin stream

        if isinstance(track, MediaEntry):
            # single track entry (MediaEntry)
            self.now_playing.update(track)

            # update playlist position
            if idx > -1:
                self.playlist.set_position(idx)
            # add to "now playing" if it's a new track
            elif track not in self.playlist:  # compared by uri
                self.playlist.add_entry(track)
                self.playlist.set_position(len(self.playlist) - 1)
            # find equivalent track position in playlist
            else:
                self.playlist.goto_track(track)

        elif isinstance(track, Playlist):
            # this is a playlist result (list of dicts)
            self.playlist.clear()
            for entry in track:
                self.playlist.add_entry(entry)

            # mew playlist -> reset playlist position to the start
            self.playlist.set_position(0)

            # update self.now_playing
            if len(self.playlist):
                track = self.playlist[0]
                return self.set_now_playing(track)

            # If there's no URI, the skill might be handling playback so
            # now_playing should still be updated
            self.now_playing.update(self.playlist.as_dict)

        # update gui values
        self.gui.update_current_track()
        self.gui.update_playlist()
        if self.mpris:
            self.mpris.update_props(
                {"Metadata": self.now_playing.mpris_metadata}
            )

    # stream handling
    def validate_stream(self) -> bool:
        """
        Validate that self.now_playing is playable and update the GUI if it is
        @return: True if the `now_playing` stream can be handled
        """
        if self.active_backend == PlaybackType.AUDIO:
            self.now_playing.playback = PlaybackType.AUDIO_SERVICE

        # force playback to use audio only if configured to do so
        elif self.active_backend in [PlaybackType.VIDEO] and (
                self.settings.get("force_audioservice", False) or \
                self.settings.get("playback_mode") == PlaybackMode.FORCE_AUDIOSERVICE):
            LOG.debug("Casting PlaybackType.VIDEO to PlaybackType.AUDIO_SERVICE")
            self.now_playing.playback = PlaybackType.AUDIO_SERVICE

        if not self.now_playing.uri:
            return False
        self.gui["title"] = f"Extracting info from: {self.now_playing.uri}" # will be reset in self.gui.update_current_track()
        self.now_playing.extract_stream()
        self.gui["stream"] = self.now_playing.uri
        self.gui.update_current_track()
        return True

    def on_invalid_media(self):
        """
        Handle media playback errors. Show an error and play the next track.
        """
        LOG.warning(f"Failed to play: {self.now_playing}")
        self.gui.show_playback_error()
        self.play_next()

    # media controls
    def play_media(self, track: Union[dict, MediaEntry, PluginStream, Playlist],
                   disambiguation: List[Union[dict, MediaEntry]] = None,
                   playlist: List[Union[dict, MediaEntry]] = None):
        """
        Start playing the requested media, replacing any current playback.
        @param track: dict or MediaEntry to start playing
        @param disambiguation: list of tracks returned from search
        @param playlist: list of tracks in the current playlist
        """
        if isinstance(track, dict):
            track = dict2entry(track)
        if not isinstance(track, (MediaEntry, Playlist, PluginStream)):
            raise TypeError(f"Expected MediaEntry/Playlist, got: {track}")
        if isinstance(track, Playlist) and not playlist:
            playlist = track
            track = playlist[0]
        if self.mpris:
            self.mpris.stop()
        if self.state == PlayerState.PLAYING:
            self.pause()  # make it more responsive
        if disambiguation:
            self.media.search_playlist.replace(disambiguation)
            self.media.search_playlist.sort_by_conf()
            self.gui.update_search_results()
        if playlist:
            self.playlist.replace(playlist)
        if track in self.playlist:
            self.playlist.goto_track(track)
        self.set_now_playing(track)
        self.play()

    @property
    def audio_service_player(self) -> str:
        """
        Return the configured audio player that is handling playback
        """
        if not self._audio_backend:
            self._audio_backend = self._get_preferred_audio_backend()
        return self._audio_backend

    def _get_preferred_audio_backend(self):
        """
        Check configuration and available backends to select a preferred backend

        NOTE - the bus api tells us what backends are loaded,however it does not
        provide "type", so we need to get that from config we still hit the
        messagebus to account for loading failures, even if config claims
        backend is enabled it might not load
        """
        cfg = Configuration()["Audio"]["backends"]
        available = [k for k, v in cfg.items()
                     if v.get("type", "") != "ovos_common_play"]
        preferred = self.settings.get("preferred_audio_services") or \
                    ["vlc", "simple"]
        for b in preferred:
            if b in available:
                return b
        LOG.error("Preferred audio service backend not installed")
        return "simple"

    def play(self):
        """
        Start playback of the current `now_playing` MediaEntry. Displays the GUI
        player, updates track history, emits events for any listeners, and
        updates mpris (if configured).
        """
        # stop any external media players
        if self.mpris and not self.mpris.stop_event.is_set():
            LOG.info("Requested playback with mpris not stopped")
            self.mpris.stop()
        # validate new stream
        # TODO buffering animation ?
        if not self.validate_stream():
            LOG.warning("Stream Validation Failed")
            self.on_invalid_media()
            return
        self.gui.show_player()

        self.track_history.setdefault(self.now_playing.uri, 0)
        self.track_history[self.now_playing.uri] += 1
        msg = dig_for_message() or Message("")
        LOG.debug(f"Requesting playback: {repr(self.active_backend)}")
        if self.active_backend in [PlaybackType.AUDIO_SERVICE, PlaybackType.AUDIO]:
            LOG.debug("Handling playback via audio_service")
            # we explicitly want to use an audio backend for audio only output
            self.audio_service.play([self.now_playing.uri],
                                    utterance=self.audio_service_player)
            self.bus.emit(msg.reply("ovos.common_play.track.state", {
                "state": TrackState.PLAYING_AUDIOSERVICE}))
            self.set_player_state(PlayerState.PLAYING)
        elif self.active_backend == PlaybackType.SKILL:
            LOG.debug("Requesting playback: PlaybackType.SKILL")
            self.bus.emit(msg.forward(
                f'ovos.common_play.{self.now_playing.skill_id}.play',
                self.now_playing.info))
            self.bus.emit(msg.reply("ovos.common_play.track.state", {
                "state": TrackState.PLAYING_SKILL}))
        elif self.active_backend == PlaybackType.VIDEO:
            LOG.debug("Requesting playback: PlaybackType.VIDEO")
            # handle video natively in mycroft-gui
            self.bus.emit(msg.forward("gui.player.media.service.play", {
                "track": self.now_playing.uri,
                "mime": self.now_playing.mimetype,
                "repeat": False}))
            self.bus.emit(msg.reply("ovos.common_play.track.state", {
                "state": TrackState.PLAYING_VIDEO}))
        elif self.active_backend == PlaybackType.WEBVIEW:
            LOG.debug("Requesting playback: PlaybackType.WEBVIEW")
            # open a url in native webview in mycroft-gui
            self.bus.emit(msg.reply("ovos.common_play.track.state", {
                "state": TrackState.PLAYING_WEBVIEW}))
        else:
            raise ValueError("invalid playback request")
        if self.mpris:
            self.mpris.update_props({"CanGoNext": self.can_next})
            self.mpris.update_props({"CanGoPrevious": self.can_prev})
        LOG.debug(f"self.active_backend={repr(self.active_backend)}")

    def play_shuffle(self):
        """
        Go to a random position in the playlist and set that MediaEntry as
        'now_playing` (does NOT call 'play').
        """
        LOG.debug("Shuffle == True")
        if len(self.playlist) > 1 and not self.playlist.is_last_track:
            # TODO: does the 'last track' matter in this case?
            self.playlist.set_position(random.randint(0, len(self.playlist)))
            self.set_now_playing(self.playlist.current_track)
        else:
            self.media.search_playlist.next_track()
            self.set_now_playing(self.media.search_playlist.current_track)

    def play_next(self):
        """
        Play the next track in the playlist or search results.
        End playback if there is no next track, accounting for repeat and
        shuffle settings.
        """
        if self.active_backend == PlaybackType.UNDEFINED:
            LOG.error("self.active_backend is undefined")
        elif self.active_backend in [PlaybackType.MPRIS]:
            if self.mpris:
                self.mpris.play_next()
            return
        elif self.active_backend in [PlaybackType.SKILL]:
            LOG.debug(f"Defer playing next track to skill")
            self.bus.emit(Message(
                f'ovos.common_play.{self.now_playing.skill_id}.next'))
            return
        self.pause()  # make more responsive

        if self.loop_state == LoopState.REPEAT_TRACK:
            LOG.debug("Repeating single track")
        elif self.shuffle:
            LOG.debug("Shuffling")
            self.play_shuffle()
        elif not self.playlist.is_last_track:
            self.playlist.next_track()
            self.set_now_playing(self.playlist.current_track)
            LOG.info(f"Next track index: {self.playlist.position}")
        elif not self.media.search_playlist.is_last_track and \
                self.settings.get("merge_search", True):
            while self.media.search_playlist.current_track in self.playlist:
                # Don't play media already played from the playlist
                self.media.search_playlist.next_track()
            self.set_now_playing(self.media.search_playlist.current_track)
            LOG.info(f"Next search index: "
                     f"{self.media.search_playlist.position}")
        else:
            if self.loop_state == LoopState.REPEAT and len(self.playlist):
                LOG.info("end of playlist, repeat == True")
                self.playlist.set_position(0)
            else:
                LOG.info("requested next, but there aren't any more tracks")
                self.gui.handle_end_of_playback()
                return
        self.play()

    def play_prev(self):
        """
        Play the previous track in the playlist.
        If there is no previous track, do nothing.
        """
        if self.active_backend in [PlaybackType.MPRIS]:
            if self.mpris:
                self.mpris.play_prev()
            return
        elif self.active_backend in [PlaybackType.SKILL,
                                     PlaybackType.UNDEFINED]:
            self.bus.emit(Message(
                f'ovos.common_play.{self.now_playing.skill_id}.prev'))
            return
        self.pause()  # make more responsive

        if self.shuffle:
            # TODO: Should skipping back get a random track instead of previous?
            self.play_shuffle()
        elif not self.playlist.is_first_track:
            self.playlist.prev_track()
            self.set_now_playing(self.playlist.current_track)
            LOG.debug(f"Previous track index: {self.playlist.position}")
            self.play()
        else:
            LOG.debug("requested previous, but already in 1st track")

    def pause(self):
        """
        Ask the current playback to pause.
        """
        LOG.debug(f"Pausing playback: {self.active_backend}")
        if self.active_backend in [PlaybackType.AUDIO_SERVICE,
                                   PlaybackType.UNDEFINED]:
            self.audio_service.pause()
        if self.active_backend in [PlaybackType.AUDIO,
                                   PlaybackType.VIDEO,
                                   PlaybackType.UNDEFINED]:
            self.bus.emit(Message("gui.player.media.service.pause"))
        if self.active_backend in [PlaybackType.SKILL,
                                   PlaybackType.UNDEFINED]:
            self.bus.emit(Message(f'ovos.common_play'
                                  f'.{self.active_skill}.pause'))
        if self.active_backend in [PlaybackType.MPRIS] and self.mpris:
            self.mpris.pause()
        self.set_player_state(PlayerState.PAUSED)
        self._paused_on_duck = False

    def resume(self):
        """
        Ask any paused or stopped playback to resume.
        """
        LOG.debug(f"Resuming playback: {self.active_backend}")
        if self.active_backend in [PlaybackType.AUDIO_SERVICE,
                                   PlaybackType.UNDEFINED]:
            self.audio_service.resume()

        if self.active_backend in [PlaybackType.SKILL,
                                   PlaybackType.UNDEFINED]:
            self.bus.emit(
                Message(f'ovos.common_play.{self.active_skill}.resume'))

        if self.active_backend in [PlaybackType.AUDIO,
                                   PlaybackType.VIDEO]:
            self.bus.emit(Message('gui.player.media.service.resume'))

        if self.active_backend in [PlaybackType.MPRIS] and self.mpris:
            self.mpris.resume()

        self.set_player_state(PlayerState.PLAYING)

    def seek(self, position: int):
        """
        Request playback to go to a specific position in the current media
        @param position: milliseconds position to seek to
        """
        if self.active_backend in [PlaybackType.AUDIO_SERVICE,
                                   PlaybackType.UNDEFINED]:
            self.audio_service.set_track_position(position / 1000)
        self.gui["position"] = position

    def stop(self):
        """
        Request stopping current playback and searching
        """
        # stop any search still happening
        self.bus.emit(Message("ovos.common_play.search.stop"))

        LOG.debug("clearing playlist")
        self.playlist.clear()  # needed to ensure next track doesnt track due to autoplay

        LOG.debug("Stopping playback")
        if self.active_backend in [PlaybackType.AUDIO_SERVICE,
                                   PlaybackType.UNDEFINED]:
            self.stop_audio_service()
            self.set_player_state(PlayerState.STOPPED)
        if self.active_backend in [PlaybackType.SKILL,
                                   PlaybackType.UNDEFINED]:
            self.stop_audio_skill()
        if self.active_backend in [PlaybackType.AUDIO,
                                   PlaybackType.VIDEO,
                                   PlaybackType.UNDEFINED]:
            self.stop_gui_player()
            self.set_player_state(PlayerState.STOPPED)
        # if self.active_backend in [PlaybackType.MPRIS] and self.mpris:
        #    self.mpris.stop()

    def stop_gui_player(self):
        """
        Emit a Message notifying the gui player to stop
        """
        self.bus.emit(Message("gui.player.media.service.stop"))

    def stop_audio_skill(self):
        """
        Emit a Message notifying self.active_skill to stop
        """
        self.bus.emit(Message(f'ovos.common_play.{self.active_skill}.stop'))

    def stop_audio_service(self):
        """
        Call self.audio_service.stop()
        """
        self.audio_service.stop()

    def reset(self):
        """
        Reset this instance to clear any media or settings
        """
        self.stop()
        self.playlist.clear()
        self.media.search_playlist.clear()
        self.set_media_state(MediaState.NO_MEDIA)
        self.shuffle = False
        self.loop_state = LoopState.NONE

    def shutdown(self):
        """
        Shutdown this instance and its spawned objects. Remove events.
        """
        self.stop()
        if self.mpris:
            self.mpris.shutdown()
        self.now_playing.shutdown()
        self.gui.shutdown()
        self.media.shutdown()
        self.remove_event('recognizer_loop:record_begin')
        self.remove_event('recognizer_loop:record_end')
        self.remove_event('gui.player.media.service.sync.status')
        self.remove_event("gui.player.media.service.get.next")
        self.remove_event("gui.player.media.service.get.previous")

    # player -> common play
    @require_default_session()
    def handle_player_state_update(self, message):
        """
        Handles 'gui.player.media.service.sync.status' and
        'ovos.common_play.player.state' messages with player state updates
        @param message: Message providing new "state" data
        """
        state = message.data.get("state")
        if state is None:
            raise ValueError(f"Got state update message with no state: "
                             f"{message}")
        if isinstance(state, int):
            state = PlayerState(state)
        if not isinstance(state, PlayerState):
            raise ValueError(f"Expected int or PlayerState, but got: {state}")
        if state == self.state:
            return
        LOG.info(f"PlayerState changed: {repr(state)}")
        if state == PlayerState.PLAYING:
            self.state = PlayerState.PLAYING
        elif state == PlayerState.PAUSED:
            self.state = PlayerState.PAUSED
            if self.app_view_timeout_enabled and \
                    self.app_view_timeout_mode == "pause":
                LOG.debug("Starting GUI pause timeout counter")
                self.gui.cancel_app_view_timeout()
                self.gui.schedule_app_view_pause_timeout()
        elif state == PlayerState.STOPPED:
            self.state = PlayerState.STOPPED

        if self.mpris:
            state2str = {PlayerState.PLAYING: "Playing",
                         PlayerState.PAUSED: "Paused",
                         PlayerState.STOPPED: "Stopped"}
            self.mpris.update_props({"CanPause": state == PlayerState.PLAYING,
                                     "CanPlay": state == PlayerState.PAUSED,
                                     "PlaybackStatus": state2str[state]})

    @require_default_session()
    def handle_player_media_update(self, message):
        """
        Handles 'ovos.common_play.media.state' messages with media state updates
        @param message: Message providing new "state" data
        """
        LOG.debug(f"backend={repr(self.active_backend)}|"
                  f"msg_type={message.msg_type}")
        state = message.data.get("state")
        if state is None:
            raise ValueError(f"Got state update message with no state: "
                             f"{message}")
        if isinstance(state, int):
            state = MediaState(state)
        if not isinstance(state, MediaState):
            raise ValueError(f"Expected int or MediaState, but got: {state}")
        if state == self.media_state:
            return
        LOG.debug(f"MediaState changed: {repr(state)}")
        self.media_state = state
        if state == MediaState.END_OF_MEDIA:
            self.handle_playback_ended(message)
        elif state == MediaState.INVALID_MEDIA:
            self.handle_invalid_media(message)
            if self.settings.get("autoplay", True):
                self.play_next()

    @require_default_session()
    def handle_invalid_media(self, message):
        self.gui.show_playback_error()

    @require_default_session()
    def handle_playback_ended(self, message):
        # TODO: When we get here, self.active_backend has been reset!
        LOG.info(f"END OF MEDIA - playlist pos: {self.playlist.position} "
                  f"total tracks: {len(self.playlist)} "
                 f"backend: {self.active_backend}")
        go_next = self.settings.get("autoplay", True) and \
                self.active_backend != PlaybackType.MPRIS and \
                self.playlist.position + 1 < len(self.playlist)
        LOG.debug(f"Go to Next track: {go_next}")
        if go_next:
            self.play_next()
            return
        LOG.info("Playback ended")
        self.gui.handle_end_of_playback(message)

    # ovos common play bus api requests
    @require_default_session()
    def handle_search_replace(self, message):
        LOG.debug("Updating search results playlist")
        pl = message.data["playlist"]
        replace = message.data.get("replace", False)
        sort = message.data.get("sort_by_conf", True)
        if replace:
            if not pl:
                self.media.search_playlist.clear()
            else:
                self.media.search_playlist.replace(pl)
        else:
            for e in pl:
                self.media.search_playlist.add_entry(e)
        if sort:
            self.media.search_playlist.sort_by_conf()
        self.gui.update_search_results()

    @require_default_session()
    def handle_play_request(self, message):
        LOG.debug("Received playback request")
        repeat = message.data.get("repeat", False)
        if repeat:
            self.loop_state = LoopState.REPEAT
        media = message.data.get("media")
        playlist = message.data.get("playlist") or [media]
        disambiguation = message.data.get("disambiguation") or []
        self.play_media(media, disambiguation, playlist)

    @require_default_session()
    def handle_pause_request(self, message):
        self.pause()

    @require_default_session()
    def handle_stop_request(self, message):
        self.stop()

    @require_default_session()
    def handle_resume_request(self, message):
        self.resume()

    @require_default_session()
    def handle_seek_request(self, message):
        # from bus api
        miliseconds = message.data.get("seconds", 0) * 1000

        # from audio player GUI
        position = message.data.get("seekValue")
        if not position:
            position = self.now_playing.position or 0
            if self.active_backend in [PlaybackType.AUDIO_SERVICE,
                                       PlaybackType.UNDEFINED]:
                position = self.audio_service.get_track_position() or position
            position += miliseconds
        self.seek(position)

    @require_default_session()
    def handle_next_request(self, message):
        self.play_next()

    @require_default_session()
    def handle_prev_request(self, message):
        self.play_prev()

    @require_default_session()
    def handle_set_shuffle(self, message):
        self.shuffle = True
        self.gui.update_seekbar_capabilities()

    @require_default_session()
    def handle_unset_shuffle(self, message):
        self.shuffle = False
        self.gui.update_seekbar_capabilities()

    @require_default_session()
    def handle_set_repeat(self, message):
        self.loop_state = LoopState.REPEAT
        self.gui.update_seekbar_capabilities()

    @require_default_session()
    def handle_unset_repeat(self, message):
        self.loop_state = LoopState.NONE
        self.gui.update_seekbar_capabilities()

    # playlist control bus api
    @require_default_session()
    def handle_repeat_toggle_request(self, message):
        if self.loop_state == LoopState.REPEAT_TRACK:
            self.loop_state = LoopState.NONE
        elif self.loop_state == LoopState.REPEAT:
            self.loop_state = LoopState.REPEAT_TRACK
        elif self.loop_state == LoopState.NONE:
            self.loop_state = LoopState.REPEAT
        LOG.info(f"Repeat: {self.loop_state}")
        self.gui.update_seekbar_capabilities()

    @require_default_session()
    def handle_shuffle_toggle_request(self, message):
        self.shuffle = not self.shuffle
        LOG.info(f"Shuffle: {self.shuffle}")
        self.gui.update_seekbar_capabilities()

    @require_default_session()
    def handle_playlist_set_request(self, message):
        self.playlist.clear()
        self.handle_playlist_queue_request(message)

    @require_default_session()
    def handle_playlist_queue_request(self, message):
        for track in message.data["tracks"]:
            self.playlist.add_entry(track)
        self.gui.update_playlist()

    @require_default_session()
    def handle_playlist_clear_request(self, message):
        self.playlist.clear()
        self.set_media_state(MediaState.NO_MEDIA)
        self.gui.update_playlist()

    # audio ducking
    @require_default_session()
    def handle_duck_request(self, message):
        """
        Pause audio on 'recognizer_loop:record_begin'
        @param message: Message associated with event
        """
        if self.state == PlayerState.PLAYING:
            self.pause()
            self._paused_on_duck = True

    @require_default_session()
    def handle_unduck_request(self, message):
        """
        Resume paused audio on 'recognizer_loop:record_begin'
        @param message: Message associated with event
        """
        if self.state == PlayerState.PAUSED and self._paused_on_duck:
            self.resume()
            self._paused_on_duck = False

    # track data
    @require_default_session()
    def handle_track_length_request(self, message):
        l = self.now_playing.length
        if self.active_backend == PlaybackType.AUDIO_SERVICE:
            l = self.audio_service.get_track_length() or l
        data = {"length": l}
        self.bus.emit(message.response(data))

    @require_default_session()
    def handle_track_position_request(self, message):
        pos = self.now_playing.position
        if self.active_backend == PlaybackType.AUDIO_SERVICE:
            pos = self.audio_service.get_track_position() or pos
        data = {"position": pos}
        self.bus.emit(message.response(data))

    @require_default_session()
    def handle_set_track_position_request(self, message):
        miliseconds = message.data.get("position")
        self.seek(miliseconds)

    @require_default_session()
    def handle_track_info_request(self, message):
        data = self.now_playing.as_dict
        if self.active_backend == PlaybackType.AUDIO_SERVICE:
            data = self.audio_service.track_info() or data
        self.bus.emit(message.response(data))

    # internal info
    @require_default_session()
    def handle_list_backends_request(self, message):
        data = self.audio_service.available_backends()
        self.bus.emit(message.response(data))

    # app timeout
    @property
    def app_view_timeout_enabled(self):
        return self.settings.get("app_view_timeout_enabled", False)

    @property
    def app_view_timeout_value(self):
        return self.settings.get("app_view_timeout", 30)

    @property
    def app_view_timeout_mode(self):
        return self.settings.get("app_view_timeout_mode", "all")

    @require_default_session()
    def handle_enable_app_timeout(self, message):
        self.settings["app_view_timeout_enabled"] = message.data.get("enabled", False)
        self.settings.store()
        if not self.app_view_timeout_enabled:
            self.gui.cancel_app_view_timeout()

    @require_default_session()
    def handle_set_app_timeout(self, message):
        # timeout in seconds: 15 | 30 | 45 | 60
        self.settings["app_view_timeout"] = message.data.get("timeout", 30)
        self.settings.store()
        self.gui.cancel_app_view_timeout(restart=True)

    @require_default_session()
    def handle_set_app_timeout_mode(self, message):
        # timeout modes: all | pause
        self.settings["app_view_timeout_mode"] = message.data.get("mode", "all")
        self.settings.store()
        self.gui["app_view_timeout_mode"] = self.settings.get("app_view_timeout_mode", "all")
        self.gui.cancel_app_view_timeout()



from pprint import pformat

from ovos_bus_client import Message
from ovos_utils.log import LOG
from ovos_workshop.decorators.ocp import *

from ovos_plugin_common_play.ocp import OCP
from ovos_plugin_common_play.ocp.base import OCPAudioPlayerBackend
from ovos_plugin_common_play.ocp.constants import OCP_ID


class OCPAudioBackend(OCPAudioPlayerBackend):
    """ This plugin makes regular mycroft skills that use the audio service
    go trough the OVOS common play framework, this plugin simply delegates
    the task by emitting bus messages expected by Ovos Common Play API. If
    configured it will also launch OCP itself """

    def __init__(self, config, bus=None, name='ovos.common_play'):
        super(OCPAudioBackend, self).__init__(config=config,
                                              bus=bus)
        self.ocp = None
        self.name = name
        self._track_info = {}
        self.bus.on("gui.player.media.service.set.meta",
                    self.handle_receive_meta)
        self.create_ocp(self.config)

    @property
    def player(self):
        return self.ocp.player

    def create_ocp(self, config: dict):
        self.config = config
        self.config.setdefault("mode", "auto")
        LOG.debug(f"OCP config:\n {pformat(config)}")

        if self.config["mode"] == "external":
            # flag for external OCP, eg, system service daemon
            # send only bus messages and don't create ocp object
            self.ocp = None
        elif self.config["mode"] == "auto":
            # if OCP is already running connect to it, else launch it
            if not self.bus.wait_for_response(Message("ovos.common_play.ping"),
                                              "ovos.common_play.pong"):
                try:
                    self.ocp = OCP(bus=self.bus, settings=self.config,
                                   skill_id=OCP_ID)
                except Exception as e:
                    # otherwise stack trace is swallowed by plugin loader
                    LOG.exception(e)
                    raise
        else:
            try:
                self.ocp = OCP(bus=self.bus, settings=self.config,
                               skill_id=OCP_ID)
            except Exception as e:
                # otherwise stack trace is swallowed by plugin loader
                LOG.exception(e)
                raise

    def handle_receive_meta(self, message):
        self._track_info = message.data

    def supported_uris(self):
        return ['file', 'http', 'https']

    def play(self, repeat=False):
        """ Play media playback. """
        # self._tracks is populated in base class this is derived from
        if len(self._tracks):
            self._track_info = self._tracks[0]
            self.bus.emit(Message('ovos.common_play.play',
                                  {'repeat': repeat,
                                   "media": self._tracks[0],
                                   "playlist": self._tracks}))

    def pause(self):
        self.bus.emit(Message("ovos.common_play.pause"))

    def resume(self):
        self.bus.emit(Message("ovos.common_play.resume"))

    def stop(self):
        self._track_info = {}
        self.bus.emit(Message("ovos.common_play.stop"))
        if self.ocp:
            return True
        return False

    def get_track_length(self):
        """
        getting the duration of the audio in milliseconds
        NOTE: not yet supported by mycroft-core
        """
        length = 0
        msg = Message('ovos.common_play.get_track_length')
        info = self.bus.wait_for_response(msg, timeout=1)
        if info:
            length = info.data.get("length", 0)
        return length

    def get_track_position(self):
        """
        get current position in milliseconds
        NOTE: not yet supported by mycroft-core
        """
        pos = 0
        msg = Message('ovos.common_play.get_track_position')
        info = self.bus.wait_for_response(msg, timeout=1)
        if info:
            pos = info.data.get("position", 0)
        return pos

    def set_track_position(self, milliseconds):
        """
        go to position in milliseconds
        NOTE: not yet supported by mycroft-core
          Args:
                milliseconds (int): number of milliseconds of final position
        """
        msg = Message('ovos.common_play.set_track_position',
                      {"position": milliseconds})
        self.bus.emit(msg)

    def seek_forward(self, seconds=1):
        """Skip X seconds.

        Arguments:
            seconds (int): number of seconds to seek, if negative rewind
        """
        msg = Message('ovos.common_play.seek', {"seconds": seconds})
        self.bus.emit(msg)

    def seek_backward(self, seconds=1):
        """Rewind X seconds.

        Arguments:
            seconds (int): number of seconds to seek, if negative jump forward.
        """
        msg = Message('ovos.common_play.seek', {"seconds": seconds * -1})
        self.bus.emit(msg)

    def track_info(self):
        """
            Fetch info about current playing track.
            Returns:
                Dict with track info.
        """
        if not self._track_info and self.ocp:
            return self.ocp.player.now_playing.as_dict
        return self._track_info

    def shutdown(self):
        self.bus.remove("gui.player.media.service.set.meta",
                        self.handle_receive_meta)
        if self.ocp is not None:
            self.ocp.shutdown()


def load_service(base_config, bus):
    backends = base_config.get('backends', [])
    services = [(b, backends[b]) for b in backends
                if backends[b]['type'] == 'ovos_common_play' and
                backends[b].get('active', True)]
    instances = [OCPAudioBackend(s[1], bus, s[0]) for s in
                 services]
    return instances


OCPPluginConfig = {
    "local": {
        "type": "ovos_common_play",
        "active": True,
        # all values below are optional

        # plugin config
        # operational mode refers to the OCP integration
        # "external" - OCP is already running elsewhere, connect by bus only
        # "native" - launch OCP service from the plugin
        # "auto" - if OCP is already running connect to it, else launch it
        # you should only change this if you want to run OCP as a standalone system service
        "mode": "auto",

        # MPRIS integrations
        # integration is enabled by default, but can be disabled
        "disable_mpris": False,
        # dbus type for MPRIS, "session" or "system"
        "dbus_type": "session",
        # allow OCP to control MPRIS enabled 3rd party applications
        # voice enable them (next/prev/stop/resume..)
        # and stop them when OCP starts it's own playback
        # NOTE: OCP can be controlled itself via MPRIS independentely of this setting
        "manage_external_players": False,

        # Playback settings
        # AUTO = 0 - play each entry as considered appropriate,
        #            ie, make it happen the best way possible
        # AUDIO_ONLY = 10  - only consider audio entries
        # VIDEO_ONLY = 20  - only consider video entries
        # FORCE_AUDIO = 30 - cast video to audio unconditionally
        #                   (audio can still play in mycroft-gui)
        # FORCE_AUDIOSERVICE = 40 - cast everything to audio service backend,
        #                           mycroft-gui will not be used
        # EVENTS_ONLY = 50 - only emit ocp events,
        #                    do not display or play anything.
        #                    allows integration with external interfaces
        "playback_mode": 0,

        # ordered list of audio backend preferences,
        # when OCP selects a audio service for playback
        # this list is checked in order until a available backend is found
        "preferred_audio_services": ["vlc", "mplayer", "simple"],

        # to handle video playback different backends are supported
        #   - valid options - "native", "qtav", "auto"
        # "qtav" does not support widget integration
        # "native" sometimes has problems playing stream urls
        "video_player_backend": "auto",

        ## when media playback ends "click next"
        "autoplay": True,
        # if True behaves as if the search results are part of the playlist
        #  eg:
        #   - click next in last track -> play next search result
        #   - end of playlist + autoplay -> play next search result
        "merge_search": True,

        # search params
        # minimum time to wait for skill replies,
        # after this time, if at least 1 result was
        # found, selection is triggered
        "min_timeout": 5,
        # maximum time to wait for skill replies,
        # after this time, regardless of number of
        # results, selection is triggered
        "max_timeout": 15,
        # ignore results below min_Score
        "min_score": 50,
        # stop collecting results if we get a
        # match with confidence >= early_stop_thresh
        "early_stop_thresh": 85,
        # sleep this amount before early stop,
        # allows skills that "just miss" to also be taken into account
        "early_stop_grace_period": 0.5,
        # if True emits the regular mycroft-core
        # bus messages to get results from "old style" skills
        "backwards_compatibility": True,
        # allow skills to request more time,
        # extends min_timeout for individual queries (up to max_timeout)
        "allow_extensions": True,
        # if no results for a MediaType, perform a second query with MediaType.GENERIC
        "search_fallback": True,

        # stream extractor settings
        # how to handle bandcamp streams
        # "pybandcamp", "youtube-dl"
        "bandcamp_backend": "pybandcamp",
        # how to handle youtube streams
        # "youtube-dl", "pytube", "pafy", "invidious"
        "youtube_backend": "invidious",
        # the url to the invidious instance to be used"
        # by default uses a random instance
        "invidious_host": None,
        # get final stream locally or from where invidious is hosted
        # This partially allows bypassing geoblocked content,
        # but it is a global flag, not per entry.
        "invidious_proxy": False,
        # different forks of youtube-dl are supported
        # "yt-dlp", "youtube-dl", "youtube-dlc"
        "ydl_backend": "yt-dlp",
        # how to extract live streams from a youtube channel
        # "pytube", "youtube_searcher", "redirect", "youtube-dl"
        "youtube_live_backend": "redirect"  # uses youtube auto redirect https://www.youtube.com/{channel_name}/live
    }
}
